/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *   LCD key support added by Mike Willett <mike@digital-chaos.co.uk>
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

// Compile only, if this driver is enabled
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef AUDIOBOX_ENABLE_LCDPROC

#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "socket.h"
#include "keytab.h"
#include "display_lcdproc.h"

// These includes provide the context object definitions
//
#include "song.h"
#include "info.h"
#include "search.h"
#include "menu.h"
#include "main.h"

#include "util.h"

/******************************************************************************/

#ifndef LCDPORT
#  define LCDPORT 13666
#endif


/******************************************************************************/

static unsigned int scroll_next;	// time for the next scroll event
static int scroll_delta	= 1;		// 1 or -1, will be added to idx
static int scroll_idx	= 0;		// scroll index

/******************************************************************************/

// these macros need some reworking... we really don't need both of them.
//
#define xy_printf(_col,_line,_arg...)           \
    {   \
        char _buf[MAX_PATH_LEN];    \
        char _cmd[MAX_PATH_LEN];    \
        sprintf(_buf,_arg);         \
        printf("\033[%d;%dH\033[0;42;30m%-40.40s\033[0m",_line+1+18,_col+1+20,_buf); \
	fflush(stdout); \
        sprintf(_cmd,"widget_set AUDIOBOX line%d %d %d {%s}\n",_line,_col,(_line)+1,_buf); \
        socket_send(lcd_display_fd(display), _cmd);    \
    }

#define lcd_printf(_arg...)         \
    {   \
        char buf[MAX_PATH_LEN];                         \
        sprintf(buf, _arg);                             \
        socket_send(lcd_display_fd(display), buf);      \
    }


static int lcd_display_open(display_t * display);
static void lcd_display_close(display_t * display);
static int lcd_display_fd(display_t * display);
static void lcd_display_handle_event(display_t * display);
static void lcd_display_show(display_t * display, void *ctx, unsigned int type);
static void lcd_display_cmd(display_t * display, unsigned int cmd, void *arg);


/******************************************************************************/

// This structure defines the generic LCDproc display attributes.
//
struct lcd_display_d
{
	display_t parent;

	int sock;		// socket fd
	int lines;		// number of display lines
	int columns;		// number of display columns

#define m_sock      ((struct lcd_display_d *) display)->sock
#define m_lines     ((struct lcd_display_d *) display)->lines
#define m_columns   ((struct lcd_display_d *) display)->columns
};


struct lcd_display_d lcd_display = {
	// intialization of parent 'class'
	//
	{
	      open:		lcd_display_open,
	      close:		lcd_display_close,
	      fd:		lcd_display_fd,
	      ev_handler:	lcd_display_handle_event,
	      show:		lcd_display_show,
	      cmd:		lcd_display_cmd,
	 },

	-1,			// sock
	2,			// lines
	40			// columns
};


/******************************************************************************/

static int lcd_display_open(display_t * display)
{
	const char *server = "localhost";

	int port = LCDPORT;
	int i;
	char *s;

	// set default LCD size
	m_columns = 40;
	m_lines = 2;

	// wait for the server to start up
	usleep(500000);

	m_sock = socket_connect(server, port);
	if (m_sock <= 0)
	{
		printf("Error connecting to LCD server %s on port %i.\n"
		       "Check to see that the server is running and operating normally.\n", server,
		       port);
		return -1;
	}
	socket_send(m_sock, "hello\n");
	usleep(500000);		// wait for the server to say hi.

	// Init the screen
	//
	lcd_printf("screen_add AUDIOBOX\n");
	lcd_printf("screen_set AUDIOBOX -backlight on -priority 16 -name AUDIOBOX -heartbeat off\n");

	// Add key requests for the LCD keys.
	//
	for (i = 0; i < m_lines; ++i)
	{
		lcd_printf("widget_add AUDIOBOX line%d string\n", i);
	}

//	for (s = "KBFDRLJPQ"; *s; s++)
	for (s = "123456789*0#"; *s; s++)
	{
		lcd_printf("client_add_key %c\n", *s);
	}


	return 0;
}


/******************************************************************************/

static void lcd_display_close(display_t * display)
{
	socket_close(m_sock);
}


/******************************************************************************/

static int lcd_display_fd(display_t * display)
{
	return m_sock;
}


/******************************************************************************/

static unsigned int lcd_display_translate_key(const char key)
{
	switch (key)
	{
		case '1':	return 'r';
		case '2':	return 't';
		case '3':	return 'i';
		case '4':	return KEY_DOWN;
		case '5':	return KEY_SELECT;
		case '6':	return KEY_UP;
		case '7':	return KEY_DOWN_10;
		case '8':	return KEY_RIGHT;
		case '9':	return KEY_UP_10;
		case '0':	return KEY_LEFT;
		case '#':	return KEY_ESC;
		case '*':	return ' ';
	}
	return 0;
}


/******************************************************************************/

// The token handling code below has been snipped out of the LCDproc client.
//

static void lcd_display_handle_event(display_t * display)
{
	int i;

	char buf[8192];
	char *argv[256];
	int argc;
	int newtoken;
	int len;

	// Check for server input...
	len = socket_receive(m_sock, buf, sizeof(buf));


	// Handle server input...
	while (len > 0)
	{
		// Now split the string into tokens...
		//
		argc = 0;
		newtoken = 1;
		for (i = 0; i < len; i++)
		{
			switch (buf[i])
			{
			case ' ':
				newtoken = 1;
				buf[i] = 0;
				break;
			default:	// regular chars, keep tokenizing
				if (newtoken)
				{
					argv[argc] = buf + i;
					argc++;
				}
				newtoken = 0;
				break;
			case '\0':
			case '\n':
				buf[i] = 0;
				if (argc > 0)
				{
					if (0 == strcmp(argv[0], "listen"))
					{
#ifndef NDEBUG
						printf("Listen %s\n", argv[1]);
#endif
					}
					else if (0 == strcmp(argv[0], "ignore"))
					{
						printf("Ignore %s\n", argv[1]);
					}
					else if (0 == strcmp(argv[0], "key"))
					{
#ifndef NDEBUG
						printf("Key %s\n", argv[1]);
#endif

						handle_key(
							lcd_display_translate_key(*(argv[1])));
					}
					else if (0 == strcmp(argv[0], "menu"))
					{
						printf("Menu %s\n", argv[1]);
					}
					else if (0 == strcmp(argv[0], "connect"))
					{
						int a;

						for (a = 1; a < argc; a++)
						{
							if (0 == strcmp(argv[a], "wid"))
							{
								a++;
								m_columns = atoi(argv[a]);
							}
							else if (0 == strcmp(argv[a], "hgt"))
							{
								a++;
								m_lines = atoi(argv[a]);
							}
							else if (0 == strcmp(argv[a], "cellwid"))
							{
								a++;
							}
							else if (0 == strcmp(argv[a], "cellhgt"))
							{
								a++;
							}
						}
						lcd_printf("client_set -name audioBox\n");
					}
					else if (0 == strcmp(argv[0], "bye"))
					{
						// ignore
					}
					else if (0 == strcmp(argv[0], "success"))
					{
						// ignore
					}
					else
					{
						// print any remaining
						// unhandled messages
						//
						int j;

						for (j = 0; j < argc; j++)
							printf("%s ", argv[j]);
						printf("\n");
					}
				}

				// Restart tokenizing
				argc = 0;
				newtoken = 1;
				break;
			}	// switch(buf[i])
		}

		len = socket_receive(m_sock, buf, sizeof(buf));
	}
}


/******************************************************************************
 *
 * SHOW functions for the LCD driver
 *
 ******************************************************************************/

static void lcd_display_show_menu(display_t * display, void *ctx)
{
	int i;
	int line;

	// If we have a 1 line display, we do not bother about what to display,
	// we simply display the current entry...
	//
	if (1 == m_lines)
	{
		file_entry_t *entry = list_data(list_nth(c_start, c_selection));

		if (entry)
		{
			char *name;

			name = c_view_fullpath ? entry->fullpath : entry->name;

			xy_printf(1, 0, ">%s%s",
				  FILE_TYPE_DIR == file_type(entry) ? "+" : "",
				  &name[c_left_margin]);
		}
	}

	// Check, if the selection is still within our display area
	//
	if (c_display_top_line + m_lines <= c_selection)
	{
		// adjust downwards
		//
		c_display_top_line = c_selection - (m_lines - 1);
	}
	else if (c_display_top_line > c_selection)
	{
		// adjust upwards
		//
		c_display_top_line = c_selection;
	}


	// Loop through the visible lines and print them
	//
	for (i = 0, line = c_display_top_line; i < m_lines; i++, line++)
	{
		file_entry_t *entry = list_data(list_nth(c_start, line));

		if (entry)
		{
			char *name;

			name = c_view_fullpath ? entry->fullpath : entry->name;

			xy_printf(1, i, "%c%s%s",
				  c_selection == i + c_display_top_line ?
				  entry->marked ? '#' : '>' :
				  entry->marked ? '*' : ' ',
				  FILE_TYPE_DIR == file_type(entry) ? "+" : "",
				  &name[c_left_margin]);
		}
		else
		{
			xy_printf(1, i, "-");
		}
	}
}


/******************************************************************************/

static void lcd_display_show_search(display_t * display, void *ctx)
{
	// For 1 line displays, we'll have to save some space
	//
	if (1 == m_lines)
	{
		xy_printf(1, 0, "S(%c):>%s<", SEARCH_FULL == c_search_mode ? 'f' : 's',
			  c_search_buf);
	}
	else
	{
		xy_printf(1, 0, "Search for (%s):", SEARCH_FULL == c_search_mode ? "full" : "sub");
		xy_printf(1, 1, ">%s<", c_search_buf);
	}
}


/******************************************************************************/

static void lcd_display_show_info(display_t * display, void *ctx)
{
	int line;

	for (line = 0; line < m_lines; ++line)
	{
		xy_printf(1, line, c_text[line]);
	}
}


/******************************************************************************/

static void lcd_display_show_song(display_t * display, void *ctx)
{
	unsigned int ms_now;
	const char *pause_str = "";
	unsigned int total_name_row_len;

	// We wait a couple of seconds before we start scrolling
	// the title
	//
	ms_now = get_timestamp();


	// Initialize total row len with name length
	//
	total_name_row_len = c_name_len;

	// Set up the PAUSE string, if needed
	//
	if (SONG_MODE_PAUSE == c_current_mode)
	{
		pause_str = "PAUSED ";

		// Add length of pause string
		//
		total_name_row_len += 7;
	}


	// If the song name fits into the line, we do not scroll.
	//
	if (total_name_row_len <= m_columns)
	{
		// Display the song name 
		//
		xy_printf(1, 0, "%s%s", pause_str, c_name);
	}
	else
	{
		// Check, if the next scroll event is already due.
		//
		if (ms_now > scroll_next)
		{
			// calculate the new index and the scroll direction
			//
			scroll_idx += scroll_delta;

			if (0 == scroll_idx)
			{
				// We hit the beginning of the song name
				//
				scroll_delta = 1;

				// In the case we hit the beginning of the song
				// name, we wait an additional time before we
				// start scrolling again.
				//
				scroll_next = ms_now + SONG_SCROLL_EDGE_WAIT;
			}
			else if (total_name_row_len - m_columns == scroll_idx)
			{
				// Reached the end of the song name
				//
				scroll_delta = -1;

				// In the case we hit the end of the song name,
				// we wait an additional time before we start
				// scrolling again.
				//
				scroll_next = ms_now + SONG_SCROLL_EDGE_WAIT;
			}
			else
			{
				// Set current timestamp
				//
				scroll_next = ms_now + SONG_SCROLL_DELAY;
			}
		}
		xy_printf(1, 0, "%s%s", pause_str, &c_name[scroll_idx]);
	}

	if (c_info_show)
	{
		// Show the mp3 info
		//
		xy_printf(1, 1, "%dkBit %d.%01dkHz",
			  c_bit_rate, c_sample_rate / 1000, (c_sample_rate % 1000) / 100);
	}
	else
	{
		// Show the time and position
		//
		xy_printf(1, 1, "%s %s%s%d/%d",
			  c_time_show_remaining ? c_time_remaining : c_time,
			  c_play_random ? "R" : "", 
			  c_play_loop ? "L" : "", 
			  c_file_idx_selected + 1, 
			  list_len(c_file_list));
	}
}


/******************************************************************************/

static void lcd_display_show(display_t * display, void *ctx, unsigned int type)
{
	assert(NULL != ctx);
	assert(NULL != display);

	switch (type)
	{
	case TYPE_MENU:
	case TYPE_SONGLIST_MENU:
		lcd_display_show_menu(display, ctx);
		break;
	case TYPE_SEARCH:
		lcd_display_show_search(display, ctx);
		break;
	case TYPE_INFO:
		lcd_display_show_info(display, ctx);
		break;
	case TYPE_SONG:
		lcd_display_show_song(display, ctx);
		break;

	default:
		// ignore
		;
	}
}


/******************************************************************************/

static void lcd_display_cmd(display_t * display, unsigned int cmd, void *arg)
{
	scroll_idx	= 0;
	scroll_delta	= 1;
	
	// Set the initial time it takes before the song name starts scrolling.
	//
	scroll_next	= get_timestamp() + SONG_SCROLL_INITIAL_WAIT;
}


#endif // #ifdef AUDIOBOX_ENABLE_LCDPROC
