/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "util.h"
#include "defs.h"
#include "player.h"
#include "fileutil.h"

/******************************************************************************/

enum song_mode_d
{
	STOP = 0,
	PLAY,
	PAUSE,
	ENDED,
};

static int song_mode = STOP;
static float time_elapsed;
static float time_remaining;
static int current_frame;
static int frames_remaining;
static int sample_rate;
static int bit_rate;


/******************************************************************************/

static int recv_fd;
static int send_fd;
static int mod_flac123_pid = 0;	// PID of flac123


/******************************************************************************/

// Small parts of the code of this function have been taken from the irmp3
// project.
//

static int flac123_init()
{
	fd_set fds;
	struct timeval timeout = { 5, 0 };	// timeout for flac123 hello message

	int fd_send[2];
	int fd_recv[2];
	int oldstat;

	// -y == resync when encountering broken headers. This is useful when
	// playing streams from the web.
	//
	// I don't remember why the "dummy" argument is there. It seems to work
	// without it, however there must have been some reason. Also, it does not
	// hurt --- I'll just leave it there.
	//
// -wgr- 	char *args[] = { "/usr/bin/flac123", "-y", "-R", "dummy", NULL };
	char *args[] = { "/usr/bin/flac123", "-R", NULL };

	// make socketpair to communicate with player
	//
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_send) < 0)
	{
		printf("flac123_init(): unable to setup send pipe for flac123 (%s).\n",
		       strerror(errno));
		return -1;
	}
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_recv) < 0)
	{
		printf("flac123_init(): unable to setup recv pipe for flac123 (%s).\n",
		       strerror(errno));
		return -1;
	}


	// fork child process
	mod_flac123_pid = fork();

	if (mod_flac123_pid == 0)
	{
		// pipe in/output through socket to parent
		dup2(fd_send[0], STDIN_FILENO);
		close(fd_send[0]);
		close(fd_send[1]);
		dup2(fd_recv[0], STDOUT_FILENO);
		dup2(fd_recv[0], STDERR_FILENO);
		close(fd_recv[0]);
		close(fd_recv[1]);

		// spawn player
		execvp("/usr/bin/flac123", args);

		// never reached if exec was ok
		exit(-1);
	}

	// parent continues here
	//
	close(fd_send[0]);
	close(fd_recv[0]);
	send_fd = fd_send[1];
	recv_fd = fd_recv[1];

	oldstat = fcntl(recv_fd, F_GETFL, 0);
	fcntl(recv_fd, F_SETFL, oldstat | O_NONBLOCK);

	// wait for the hello message of the flac123 process
	//
	FD_ZERO(&fds);
	FD_SET(recv_fd, &fds);
	if (0 > select(recv_fd + 1, &fds, NULL, NULL, &timeout))
	{
		// We did not get a message within the timeout.
		//
		printf("flac123_init(): player process didn't start!\n");
		return -1;
	}
	// We did get a message. We don't care what exactly it says (should be
	// the hello message). But we know the process is up...
	//
	return 0;
}


/******************************************************************************/

static int flac123_fd()
{
	return recv_fd;
}


/******************************************************************************/

static int flac123_is_playing()
{
	return PLAY == song_mode;
}


/******************************************************************************/

static int flac123_is_paused()
{
	return PAUSE == song_mode;
}


/******************************************************************************/

static int flac123_has_ended()
{
	// We reset the flag to STOP once it is polled to avoid a 'jitter'
	// situation where multiple polls will always result in ENDED.
	//
	if (ENDED == song_mode)
	{
		song_mode = STOP;
		return TRUE;
	}
	return FALSE;
}


/******************************************************************************/

static const char *flac123_time()
{
	static char song_time[TIME_STR_LEN];

	sprintf(song_time, "%01d:%02d:%02d",
		(int) time_elapsed / 3600, ((int) time_elapsed / 60) % 60, (int) time_elapsed % 60);

	return song_time;
}


/******************************************************************************/

static const char *flac123_time_remaining()
{
	static char song_time_remaining[TIME_STR_LEN];

	sprintf(song_time_remaining, "-%01d:%02d:%02d",
		(int) time_remaining / 3600,
		((int) time_remaining / 60) % 60, (int) time_remaining % 60);

	return song_time_remaining;
}


/******************************************************************************/

static int flac123_current_frame()
{
	return current_frame;
}


/******************************************************************************/

static int flac123_sample_rate()
{
	return sample_rate;
}


/******************************************************************************/

static int flac123_bit_rate()
{
	return bit_rate;
}


/******************************************************************************/

static void flac123_send(const char *cmd)
{
	ssize_t len;

	len = write(send_fd, cmd, strlen(cmd));

	// We should check len here...
}


/******************************************************************************/

static void flac123_play(const char *path)
{
	char buf[MAX_PATH_LEN];

	assert(strlen(path) + 7 < sizeof(buf));

	sprintf(buf, "LOAD %s\n", path);
	flac123_send(buf);

	song_mode = PLAY;
}


/******************************************************************************/

static int flac123_stop()
{
	if (PLAY == song_mode)
	{
		flac123_send("STOP\n");
		song_mode = STOP;
	}
	return song_mode == STOP ? 0 : -1;
}


/******************************************************************************/

static int flac123_pause()
{
	if ((PLAY == song_mode) || (PAUSE == song_mode))
	{
		song_mode = (song_mode == PAUSE) ? PLAY : PAUSE;

		flac123_send("PAUSE\n");
		return 0;
	}
	return -1;
}


/******************************************************************************/

static void flac123_fwd()
{
	flac123_send("JUMP +500\n");
}


/******************************************************************************/

static void flac123_rew()
{
	flac123_send("JUMP -500\n");
}


/******************************************************************************/

static void flac123_goto_frame(int frame)
{
	char framestr[32];

	sprintf(framestr, "JUMP %d\n", frame);
	flac123_send(framestr);
}


/******************************************************************************/

static void flac123_parse_message(const char *msg)
{
	assert(msg);

	// We assume that this is a well formed message
	//
	switch (msg[1])
	{
		// NOTE: case headers are taken from
		//       /usr/share/doc/flac123/README.remote.gz

		////////////////////////////////////////////////////////////
		// @F <a> <b> <c> <d>
		//         Status message during playing (frame info)
		//         a = framecount (int)
		//         b = frames left this song (int)
		//         c = seconds (float)
		//         d = seconds left (float)
	case 'F':
		{
			// scan elapsed and remaining time
			//
			if (4 == sscanf(msg + 3, "%d %d %f %f",
					&current_frame,
					&frames_remaining, &time_elapsed, &time_remaining))
			{
				// If there are no more frames
				// remaining for this song, we change
				// to ENDED mode to signal the end of
				// the song.
				//
				// Actually, we check for 5 remaining
				// frames, because some songs do not
				// create remaining frame messages
				// lower than 5. This means that we
				// might ignore a couple of messages of
				// songs that actually create them and
				// therefore end the song prematurely,
				// however, 5 discarded frames should
				// not be a big problem.
				//
				if (5 == frames_remaining)
				{
					song_mode = ENDED;
				}
			}
		}
		break;


		////////////////////////////////////////////////////////////
		// @I <a>
		//         Status message after loading a song (no ID3 song info)
		//         a = filename without path and extension
	case 'I':
		// not implemented yet
		break;


		////////////////////////////////////////////////////////////
		// @S <a> <b> <c> <d> <e> <f> <g> <h> <i> <j> <k> <l>
		//         Status message after loading a song (stream info)
		//         a = mpeg type (string)
		//         b = layer (int)
		//         c = sampling frequency (int)
		//         d = mode (string)
		//         e = mode extension (int)
		//         f = framesize (int)
		//         g = stereo (int)
		//         h = copyright (int)
		//         i = error protection (int)
		//         j = emphasis (int)
		//         k = bitrate (int)
		//         l = extension (int)
	case 'S':
		// just read the sample rate and bit rate for now...
		//
		{
			int dummy;
			float fdummy;
			char bufdummy[128];

			sscanf(msg + 3, "%f %d %d %s %d %d %d %d %d %d %d %d", &fdummy,	// mp3 version
			       &dummy,	// layer 1, 2, 3
			       &sample_rate,	// sample rate
			       bufdummy,	// mode string
			       &dummy,	// mode extension
			       &dummy,	// bytes per frame
			       &dummy,	// number of channels
			       &dummy,	// copyrighted?
			       &dummy,	// CRC protected
			       &dummy,	// emphasis?
			       &bit_rate,	// mp3 bit rate
			       &dummy	// extension
				);
		}
		break;

		////////////////////////////////////////////////////////////
		// @P <a>
		//         Playing status
		//         a = 0: playing stopped
		//         a = 1: playing paused
		//         a = 2: playing unpaused
	case 'P':
		switch (msg[3])
		{
			case '0': song_mode = ENDED; break;
			case '1': song_mode = PAUSE; break;
			case '2': song_mode = PLAY;  break;
			default:;	// Ignore
		}
		break;

	default:
		// silently ignore unknown message keys
		break;
	}
}


/******************************************************************************/

#define MSG_BUF_SIZE    1024

static int flac123_recv_message()
{
	char buf[MSG_BUF_SIZE];
	int bytes;

	bytes = read(recv_fd, buf, sizeof(buf) - 1);

	if (bytes > 0)
	{
		static char msg[MSG_BUF_SIZE];
		static char *pmsg;
		static char *pbuf;

		buf[bytes] = '\0';

		// Step through the individual flac123 messages.  Each statement
		// starts with a '@' character
		//
		pbuf = buf;

		while (*pbuf)
		{
			// reset msg write pointer
			pmsg = msg;

			// Search for the beginning of the next statement
			//
			while (*pbuf && ('@' != *pbuf))
				pbuf++;

			// Reached the end of the buffer?
			//
			if (!*pbuf) break;

			// Get the statement line
			//
			do
			{
				*pmsg++ = *pbuf++;
			}
			while (*pbuf && (*pbuf != '\n') && (*pbuf != '@'));

			// terminate current statement
			*pmsg = '\0';

			// and parse it
			flac123_parse_message(msg);
		}
	}
	return bytes;
}


/******************************************************************************/

static void flac123_exit()
{
	if (mod_flac123_pid)
	{
		flac123_stop();

		// Send QUIT command to player
		//
		flac123_send("Q\n");

		close(send_fd);
		close(recv_fd);

		// execute a waitpid, so we do not end up with those nasty
		// zombies...
		//
		waitpid(mod_flac123_pid, NULL, 0);

		mod_flac123_pid = 0;
	}
}


/******************************************************************************/

static player_if_t	player_reg =
{
	.init		= flac123_init,
	.exit		= flac123_exit,

	.fd		= flac123_fd,

	.play		= flac123_play,
	.stop		= flac123_stop,
	.pause		= flac123_pause,
	.fwd		= flac123_fwd,
	.rew		= flac123_rew,
	.goto_frame	= flac123_goto_frame,

	.recv_message	= flac123_recv_message,
	.current_frame	= flac123_current_frame,
	.sample_rate	= flac123_sample_rate,
	.bit_rate	= flac123_bit_rate,

	.is_playing	= flac123_is_playing,
	.is_paused	= flac123_is_paused,
	.has_ended	= flac123_has_ended,

	.time		= flac123_time,
	.time_remaining	= flac123_time_remaining,
};

static void __attribute__ ((constructor)) flac123_constructor(void)
{
	player_register(FILE_TYPE_FLAC, &player_reg);
}

