/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>

#include "list.h"


/******************************************************************************/

struct list_d
{
	struct list_d *next;	// next pointer
	struct list_d *prev;	// previous element pointer
	void *data;		// list data
};


/******************************************************************************
 *
 * list_add
 *
 * Creates a new list node, sets the data pointer and appends it to
 * the given list.
 *
 ******************************************************************************/

list_t *list_add(list_t *node, void *data)
{
	list_t *new_elem = (list_t *) malloc(sizeof(list_t));

	assert(NULL != new_elem);

	new_elem->next = NULL;
	new_elem->data = data;

	if (node)
	{
		// Elements exist in list
		//
		list_t *last = list_last(node);

		last->next = new_elem;
		new_elem->prev = last;

		return node;
	}
	else
	{
		// List is empty, this is the first element
		//
		new_elem->prev = NULL;

		return new_elem;
	}
}


/******************************************************************************
 *
 * list_delete_nth
 *
 * deletes the n-th node from the list.
 *
 * Returns the beginning of the new list.
 *
 ******************************************************************************/

list_t *list_delete_nth(list_t *list, int n)
{
	list_t *node = list_nth(list, n);

	if (list)
	{
		if (node->prev)
		{
			node->prev->next = node->next;
		}
		else
		{
			// Its the first node in the list -> adjust
			// the pointer.
			//
			list = node->next;
		}
		if (node->next)
		{
			node->next->prev = node->prev;
		}

		// free this node
		//
		free(node);
	}

	// If we just deleted the last node return a NULL list.
	//
	return list == node ? NULL : list;
}


/******************************************************************************/

list_t *list_next(list_t *node)
{
	return node ? node->next : node;
}

/******************************************************************************/


list_t *list_prev(list_t *node)
{
	return node ? node->prev : node;
}


/******************************************************************************
 *
 * Returns the data pointer of a given list node.
 *
 ******************************************************************************/

void *list_data(list_t *node)
{
	return node ? node->data : node;
}


/******************************************************************************
 *
 * list_first
 *
 * Returns the first element of a list.
 *
 ******************************************************************************/

list_t *list_first(list_t *node)
{
	if (node)
	{
		while (node->prev)
			node = node->prev;
	}

	return node;
}


/******************************************************************************
 *
 * list_last
 *
 * Returns the last element of a list.
 *
 ******************************************************************************/

list_t *list_last(list_t *node)
{
	if (node)
	{
		while (node->next)
			node = node->next;
	}
	return node;
}


/******************************************************************************
 *
 * list_copy
 *
 * Copies the given list. Note, that the list will NOT COPY AND DATA contained
 * in the list nodes, just the nodes themselves.
 *
 * After copying a list you will end up with two seperate lists pointing to 
 * the same data pointers.
 *
 ******************************************************************************/

list_t *list_copy(list_t *in)
{
	list_t *out = NULL;

	while (in)
	{
		out = list_add(out, list_data(in));
		in = list_next(in);
	}
	return out;
}


/******************************************************************************
 *
 * list_concat
 *
 * Appends list 'newlist' to list 'in'. It's OK for both lists to be NULL
 * when passed into the function.
 *
 ******************************************************************************/

list_t *list_concat(list_t *in, list_t *newlist)
{
	list_t *last = list_last(in);

	if (last)
	{
		// The in list is not empty and has a last
		// element -> append new list (if _it_ exists)
		//
		if (newlist)
		{
			last->next = newlist;
			newlist->prev = last;
		}
		return in;
	}
	else
	{
		// List is empty -> just return new list
		//
		return newlist;
	}
}


/******************************************************************************
 *
 * list_free
 *
 * This function only frees the list nodes. It does NOT DELETE ANY DATA
 * contained in the list.
 *
 ******************************************************************************/

void list_free(list_t *node)
{
	list_t *current = list_first(node);

	while (current)
	{
		list_t *next = list_next(current);

		free(current);
		current = next;
	}
}


/******************************************************************************
 *
 * list_index_of
 *
 * Returns the index of the (first) node containing the data pointer.
 *
 ******************************************************************************/

int list_index_of(list_t *list, void *data)
{
	int idx = 0;

	while (list)
	{
		if (data == list_data(list))
			return idx;

		list = list_next(list);
		idx++;
	}
	return -1;
}


/******************************************************************************
 *
 * list_nth
 *
 * Returns the n-th element of the list.
 *
 ******************************************************************************/

list_t *list_nth(list_t *list, int n)
{
	while (n-- && list)
		list = list_next(list);

	return list;
}


/******************************************************************************
 *
 * list_len
 *
 * Returns the number of list elements in the list.
 *
 ******************************************************************************/

int list_len(list_t *list)
{
	int len = 0;

	while (list)
	{
		list = list_next(list);
		len++;
	}
	return len;
}


/******************************************************************************/
