/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "util.h"
#include "defs.h"
#include "fileutil.h"
#include "player.h"

/******************************************************************************/

enum song_mode_d
{
	STOP = 0,
	PLAY,
	PAUSE,
	ENDED,
};

static int song_mode = STOP;
static float time_elapsed;
static float time_remaining;
static int current_frame;
static int sample_rate;
static int bit_rate;


/******************************************************************************/

static int recv_fd;
static int send_fd;
static int mod_ogg123_pid = 0;	// PID of ogg123


/******************************************************************************/

// Small parts of the code of this function have been taken from the irmp3
// project.
//

static int ogg123_init()
{
	fd_set fds;
	struct timeval timeout = { 5, 0 };	// timeout for ogg123 hello message

	int fd_send[2];
	int fd_recv[2];
	int oldstat;
	char *args[4] = { "/usr/bin/ogg123", "-R", "dummy", NULL };

	// make socketpair to communicate with player
	//
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_send) < 0)
	{
		printf("ogg123_init(): unable to setup send pipe for ogg123 (%s).\n",
		       strerror(errno));
		return -1;
	}
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_recv) < 0)
	{
		printf("ogg123_init(): unable to setup recv pipe for ogg123 (%s).\n",
		       strerror(errno));
		return -1;
	}


	// fork child process
	mod_ogg123_pid = fork();

	if (mod_ogg123_pid == 0)
	{
		// pipe in/output through socket to parent
		dup2(fd_send[0], STDIN_FILENO);
		close(fd_send[0]);
		close(fd_send[1]);
		dup2(fd_recv[0], STDOUT_FILENO);
		dup2(fd_recv[0], STDERR_FILENO);
		close(fd_recv[0]);
		close(fd_recv[1]);

		// spawn player
		execvp("/usr/bin/ogg123", args);

		// never reached if exec was ok
		exit(-1);
	}

	// parent continues here

	close(fd_send[0]);
	close(fd_recv[0]);
	send_fd = fd_send[1];
	recv_fd = fd_recv[1];

	oldstat = fcntl(recv_fd, F_GETFL, 0);
	fcntl(recv_fd, F_SETFL, oldstat | O_NONBLOCK);

	// wait for the hello message of the ogg123 process
	//
	FD_ZERO(&fds);
	FD_SET(recv_fd, &fds);
	if (0 > select(recv_fd + 1, &fds, NULL, NULL, &timeout))
	{
		// We did not get a message within the timeout.
		//
		printf("ogg123_init(): player process didn't start!\n");
		return -1;
	}
	// We did get a message. We don't care what exactly it says (should be
	// the hello message). But we know the process is up...
	//
	return 0;
}


/******************************************************************************/

static int ogg123_fd()
{
	return recv_fd;
}


/******************************************************************************/

static int ogg123_is_playing()
{
	return PLAY == song_mode;
}


/******************************************************************************/

static int ogg123_is_paused()
{
	return PAUSE == song_mode;
}


/******************************************************************************/

static int ogg123_has_ended()
{
	// We reset the flag to STOP once it is polled to avoid a 'jitter'
	// situation where multiple polls will always result in ENDED.
	//
	if (ENDED == song_mode)
	{
		song_mode = STOP;
		return TRUE;
	}
	return FALSE;
}


/******************************************************************************/

static const char *ogg123_time()
{
	static char song_time[TIME_STR_LEN];

	sprintf(song_time, "%01d:%02d:%02d",
		(int) time_elapsed / 3600, ((int) time_elapsed / 60) % 60, (int) time_elapsed % 60);

	return song_time;
}


/******************************************************************************/

static const char *ogg123_time_remaining()
{
	static char song_time_remaining[TIME_STR_LEN];

	sprintf(song_time_remaining, "-%01d:%02d:%02d",
		(int) time_remaining / 3600,
		((int) time_remaining / 60) % 60, (int) time_remaining % 60);

	return song_time_remaining;
}


/******************************************************************************/

static int ogg123_current_frame()
{
	return current_frame;
}


/******************************************************************************/

static int ogg123_sample_rate()
{
	return sample_rate;
}


/******************************************************************************/

static int ogg123_bit_rate()
{
	return bit_rate;
}


/******************************************************************************/

static void ogg123_send(const char *cmd)
{
	ssize_t len;

	len = write(send_fd, cmd, strlen(cmd));

	// We should check len here...
}


/******************************************************************************/

static void ogg123_play(const char *path)
{
	char buf[MAX_PATH_LEN];

	assert(strlen(path) + 7 < sizeof(buf));

	sprintf(buf, "LOAD %s\n", path);
	ogg123_send(buf);

	song_mode = PLAY;
}


/******************************************************************************/

static int ogg123_stop()
{
	if (PLAY == song_mode)
	{
		ogg123_send("STOP\n");
		song_mode = STOP;
	}
	return song_mode == STOP ? 0 : -1;
}


/******************************************************************************/

static int ogg123_pause()
{
	if ((PLAY == song_mode) || (PAUSE == song_mode))
	{
		song_mode = (song_mode == PAUSE) ? PLAY : PAUSE;

		ogg123_send("PAUSE\n");
		return 0;
	}
	return -1;
}


/******************************************************************************/

static void ogg123_fwd()
{
	// Jump ahead 10 seconds.
	//
	ogg123_send("JUMP +10\n");
}


/******************************************************************************/

static void ogg123_rew()
{
	// Jump back 10 seconds
	//
	ogg123_send("JUMP -500\n");
}


/******************************************************************************/

static void ogg123_goto_frame(int frame)
{
// ogg123 always reports current frame == 0 -> Disable goto frame.
//
// -wgr-        char framestr[32];
// -wgr- 
// -wgr-        sprintf(framestr, "JUMP %d\n", frame);
// -wgr-        ogg123_send(framestr);
}


/******************************************************************************/

static void ogg123_parse_message(const char *msg)
{
	// We assume that this is a well formed message
	//
	switch (*(msg + 1))
	{
		// @F are timestamp messages
		//
	case 'F':
		{
			int dummy;
			static int message_cnt = 0;

			// scan elapsed and remaining time
			//
			if (4 == sscanf(msg + 2, "%d %d %f %f",
					&current_frame, &dummy, &time_elapsed, &time_remaining))
			{
				// count valid @F messages only
				//
				message_cnt++;

				// If there is no more time remaining for this
				// song, we change to STOP mode to signal the
				// end of the song
				//
				// The counter helps to ignore timestamp messages
				// < 0.15 secs that are still in the receive queue.
				//
				if ((time_remaining < 0.15) && (message_cnt > 20))
				{
					song_mode = ENDED;
					message_cnt = 0;
				}
			}
		}
		break;

		// @I reports the ID3 tag
		//
	case 'I':
		// not implemented yet
		break;

		// @S reports play status
		//
	case 'S':
		// just read the sample rate for now...
		//
		{
			int dummy;
			float fdummy;
			char bufdummy[128];

			sscanf(msg + 2, "%f %d %d %s %d %d %d %d %d %d %d %d", &fdummy,	// mp3 version
			       &dummy,	// layer 1, 2, 3
			       &sample_rate,	// sample rate
			       bufdummy,	// mode string
			       &dummy,	// mode extension
			       &dummy,	// bytes per frame
			       &dummy,	// number of channels
			       &dummy,	// copyrighted?
			       &dummy,	// CRC protected
			       &dummy,	// emphasis?
			       &bit_rate,	// mp3 bit rate
			       &dummy	// extension
				);
		}
		break;

	default:
		// silently ignore unknown message keys
		break;
	}
}


/******************************************************************************/

static int ogg123_recv_message()
{
	char buf[8192];
	int bytes;

	bytes = read(recv_fd, buf, sizeof(buf) - 1);

	if (bytes > 0)
	{
		static char msg[8192];
		static char *pmsg = msg;
		static char *pbuf;

		buf[bytes] = '\0';

		// Step through the individual ogg123 messages.
		// Each statement starts with a '@' character
		//
		pbuf = buf;

		while (*pbuf)
		{
			// if we reach a '@' and already have a statement
			// in the buffer, we parse it
			//
			if (('@' == *pbuf) && (pmsg != msg))
			{
				// terminate current statement
				*pmsg = '\0';

				// and parse it
				ogg123_parse_message(msg);

				// reset msg write pointer
				pmsg = msg;
			}

			*pmsg++ = *pbuf++;
		}

	}
	return bytes;
}


/******************************************************************************/

static void ogg123_exit()
{
	if (mod_ogg123_pid)
	{
		ogg123_stop();

		// Send QUIT command to player
		//
		ogg123_send("Q\n");

		close(send_fd);
		close(recv_fd);

		send_fd = 0;
		recv_fd = 0;

		// execute a waitpid, so we do not end up with those nasty
		// zombies...
		//
		waitpid(mod_ogg123_pid, NULL, 0);

		// This is annoying, but we'll have to wait a couple of seconds
		// for ALL the ogg123 processes to quit. Otherwise, they will
		// not release the audio device and mpg123 will crash when
		// started.
		//
		sleep(2);

		mod_ogg123_pid = 0;
	}
}


/******************************************************************************/

static player_if_t	player_reg =
{
	.init		= ogg123_init,
	.exit		= ogg123_exit,

	.fd		= ogg123_fd,

	.play		= ogg123_play,
	.stop		= ogg123_stop,
	.pause		= ogg123_pause,
	.fwd		= ogg123_fwd,
	.rew		= ogg123_rew,
	.goto_frame	= ogg123_goto_frame,

	.recv_message	= ogg123_recv_message,
	.current_frame	= ogg123_current_frame,
	.sample_rate	= ogg123_sample_rate,
	.bit_rate	= ogg123_bit_rate,

	.is_playing	= ogg123_is_playing,
	.is_paused	= ogg123_is_paused,
	.has_ended	= ogg123_has_ended,

	.time		= ogg123_time,
	.time_remaining	= ogg123_time_remaining,
};

static void __attribute__ ((constructor)) ogg123_constructor(void)
{
	player_register(FILE_TYPE_OGG, &player_reg);
}

