/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "player.h"
#include "search.h"
#include "keytab.h"
#include "util.h"

// for gettimeofday
#include <sys/time.h>
#include <time.h>

#include "song.h"

/******************************************************************************/

typedef struct song_d
{
	// Screen base 'class' object
	//
	screen_t parent;

	// display context object
	//
	song_display_context_t ctx;

	// file selection history for RANDOM mode
	unsigned int file_idx_history[FILE_IDX_HISTORY_SIZE];
	unsigned int file_idx_history_idx;

	// Last song. Used for 'previous song' key
	//
	int file_idx_last_song;

	// This is the timer for ths display of the song to snap back 
	// to the currently PLAYING song. If the user selects another song,
	// it will be displayed until this timer expires.
	//
	unsigned int snap_back;

	// The information for suspending a song
	//
	int current_frame;

#define m_ctx                   ((song_t *) screen)->ctx
#define m_name                  ((song_t *) screen)->ctx.name
#define m_name_len              ((song_t *) screen)->ctx.name_len
#define m_name_show_fullpath    ((song_t *) screen)->ctx.name_show_fullpath
#define m_file_list             ((song_t *) screen)->ctx.file_list
#define m_file_idx_playing      ((song_t *) screen)->ctx.file_idx_playing
#define m_file_idx_selected     ((song_t *) screen)->ctx.file_idx_selected
#define m_current_mode          ((song_t *) screen)->ctx.current_mode
#define m_play_random           ((song_t *) screen)->ctx.play_random
#define m_play_loop             ((song_t *) screen)->ctx.play_loop
#define m_time_show_remaining   ((song_t *) screen)->ctx.time_show_remaining
#define m_info_show             ((song_t *) screen)->ctx.info_show
#define m_sample_rate           ((song_t *) screen)->ctx.sample_rate
#define m_bit_rate              ((song_t *) screen)->ctx.bit_rate
#define m_time                  ((song_t *) screen)->ctx.time
#define m_time_remaining        ((song_t *) screen)->ctx.time_remaining
#define m_file_idx_history      ((song_t *) screen)->file_idx_history
#define m_file_idx_history_idx  ((song_t *) screen)->file_idx_history_idx
#define m_file_idx_last_song    ((song_t *) screen)->file_idx_last_song
#define m_snap_back             ((song_t *) screen)->snap_back
#define m_current_frame         ((song_t *) screen)->current_frame
}
song_t;


/******************************************************************************/

// Points to the currently active SONG screen.
//
static screen_t *active_song_screen = NULL;
static int song_screen_counter = 0;

/******************************************************************************/

static int is_songfile_type(int type)
{
	return (FILE_TYPE_MP3  == type) || 
	       (FILE_TYPE_OGG  == type) || 
	       (FILE_TYPE_FLAC == type) || 
	       (FILE_TYPE_URL  == type);
}


/******************************************************************************/

screen_t *song_active_screen()
{
	return active_song_screen;
}


/******************************************************************************/

static void song_set_display_name(screen_t * screen, int idx)
{
	file_entry_t *entry;

	entry = list_data(list_nth(m_file_list, idx));
	if (entry)
	{
		save_strcpy(m_name, m_name_show_fullpath ? entry->fullpath : entry->name);
		m_name_len = strlen(m_name);
	}
	else
	{
		strcpy(m_name, "internal error!");
	}

	// Clear the display
	//
	display_cmd(CMD_CLEAR, NULL);

	// Initialize the snap back timer for the song title.
	//
	m_snap_back = get_timestamp() + SONG_SNAP_BACK_TIME;
}


/******************************************************************************/

static void song_select_prev(screen_t * screen)
{
	// RANDOM mode -> pick the previous song from the play history
	//
	if (m_play_random)
	{
		if (m_file_idx_history_idx > 0)
		{
			m_file_idx_selected = m_file_idx_history[--m_file_idx_history_idx];
		}
	}
	// No RANDOM mode -> pick previous song
	//
	else if (m_file_idx_selected > 0)
	{
		m_file_idx_selected--;
	}
	// No previous song, but LOOP mode -> wrap around
	//
	else if (m_play_loop)
	{
		m_file_idx_selected = list_len(m_file_list) -1;
	}
	// No previous song and no LOOP mode -> do nothing
	//
	else
	{
		return;
	}
	

	song_set_display_name(screen, m_file_idx_selected);
}


/******************************************************************************/

static int song_select_next(screen_t * screen)
{
	int len = list_len(m_file_list);

	// if we only have one song, we do not bother calculating the next
	// index. We indicate that there is no next song.
	//
	// UNLESS we are in loop mode. Then we return the index of the first
	// song...
	//
	if (1 == len)
	{
		if (m_play_loop)
		{
			m_file_idx_selected = 0;

			return m_file_idx_selected;
		}
		else
		{
			return -1;
		}
	}

	// Otherwise, get a new song index.
	//
	if (m_play_random)
	{
		int i;
		int idx_exists;


		// remember the current idx in the history
		//
		if (m_file_idx_history_idx < FILE_IDX_HISTORY_SIZE)
		{
			m_file_idx_history[m_file_idx_history_idx++] = m_file_idx_selected;
		}


		// If we played all the songs in the list, we say that there is
		// no more song available.
		//
		// NOTE: 
		//   m_file_idx_history_idx points to the next available entry.
		//
		// UPDATE:
		//   However, if we play in LOOP mode, we reset the history
		//   index and continue to pick a new song.
		//
		if (m_file_idx_history_idx >= len)
		{
			if (m_play_loop)
			{
				m_file_idx_history_idx = 0;
			}
			else
			{
				return -1;
			}
		}


		// Otherwise try to get a new song index.
		//
		do
		{
			m_file_idx_selected = random() % len;

			// Filter out duplicates from the file index history
			//
			idx_exists = FALSE;

			for (i = 0; i < m_file_idx_history_idx; ++i)
			{
				if (m_file_idx_selected == m_file_idx_history[i])
				{
					idx_exists = TRUE;
					break;	// for
				}
			}
		}
		while (idx_exists);
	}
	// No RANDOM mode, pick next entry in list.
	//
	else if (list_nth(m_file_list, m_file_idx_selected + 1))
	{
		m_file_idx_selected++;
	}
	// If we play in LOOP mode, we wrap around
	//
	else if (m_play_loop)
	{
		m_file_idx_selected = 0;
	}
	// No next song AND no LOOP mode.
	//
	else
	{
		// Signal end of list.
		//
		return -1;
	}

	song_set_display_name(screen, m_file_idx_selected);

	return m_file_idx_selected;
}


/******************************************************************************/

static void song_select_dir_of_current_song(screen_t * screen)
{
	file_entry_t	*fentry;
	list_t		*filelist;
	char		*path;
	char		buf[MAX_PATH_LEN];

	list_t		*iter;
	int		idx = 0;

#if 0
	screen_t *new_song;
#endif

	// Get the file entry of the song currently playing.
	//
	fentry = list_data(list_nth(m_file_list, m_file_idx_playing));

	// Get the file list of the dir
	//
	path = get_path(buf, sizeof(buf), fentry->fullpath);
	assert(NULL != path);

	filelist = scan_file_list_path(path);


	// sort it
	//
	filelist = file_entry_list_sort(filelist);


	// search for the index of the song that was currently playing
	//
	for (iter = list_first(filelist); iter; iter = list_next(iter))
	{
		file_entry_t *entry = list_data(iter);

		if (0 == strcmp(entry->fullpath, fentry->fullpath))
		{
			idx = list_index_of(filelist, entry);
			break;
		}
	}


	// First it seemed nice to have a separate screen for a new song list.
	// However, the menu with the old list still exists anyway, and it
	// seemed kind of unintuitive having the additional song screen when
	// actually using the system, therefore I removed it. If you like it
	// better with the additional screen, enable the masked out code above
	// and below.
#if 0
	// Create a new song screen with the list
	//
	new_song = song_new();
	song_set_file_list(new_song, filelist, idx);
	display_push(new_song);
#else
	song_set_file_list(screen, filelist, idx);
#endif
}


/******************************************************************************/

static void song_update(screen_t * screen)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));

	// Check, if this song has been suspended before. In this case
	// we resume playing.
	//
	if (m_current_frame)
	{
		song_resume(screen);
		m_current_frame = 0;
	}


	// The last song screen that received the show signal is
	// the active song screen
	//
	active_song_screen = screen;


	// Check, if the song mode changed to STOP. If so, play the next
	// song in the list
	//
	if (player_has_ended())
	{
		if (-1 != song_select_next(screen))
		{
			song_play(screen);
		}
		else
		{
			m_current_mode = SONG_MODE_STOP;
		}
	}
	else if (player_is_playing())
	{
		m_current_mode = SONG_MODE_PLAY;
	}
	else if (player_is_paused())
	{
		m_current_mode = SONG_MODE_PAUSE;
	}



	// Check for the song display snap back timer
	//
	if ((m_file_idx_playing != m_file_idx_selected) && (get_timestamp() > m_snap_back))
	{
		// If the snap timer expired, reset all the display
		// parameters to the currently PLAYING song.
		//
		m_file_idx_selected = m_file_idx_playing;
		song_set_display_name(screen, m_file_idx_playing);
	}


	// update the song time entries
	//
	strcpy(m_time, player_time());
	strcpy(m_time_remaining, player_time_remaining());


	// update the sample rate bit rate info
	//
	m_sample_rate = player_sample_rate();
	m_bit_rate = player_bit_rate();
}


/******************************************************************************/

static int song_handle_key(screen_t * screen, int key)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));

	switch (key)
	{
	case KEY_LEFT:
		player_rew();
		break;

	case KEY_RIGHT:
		player_fwd();
		break;

	case ' ':
		player_pause();

		// song_set_display_name() will initialize the snap
		// back timer.
		//
		song_set_display_name(screen, m_file_idx_playing);
		break;

	case KEY_BACKSPACE:
	case KEY_ESC:
		if (player_is_playing())
		{
			player_stop();
		}
		display_pop();
		break;

	case KEY_DELETE:
		// Remove the selected song from the file list
		// (not necessarily the playing song)
		//
		// Never delete the last song in the list
		//
		if (list_len(m_file_list) > 1)
		{
			m_file_list = list_delete_nth(m_file_list, m_file_idx_selected);

			// If we deleted the last song, we'll have to decrement
			// the selected idx. 
			//
			while (m_file_idx_selected > list_len(m_file_list) - 1)
			{
				m_file_idx_selected--;
			}
			song_set_display_name(screen, m_file_idx_selected);

			// Adjust the index of the currently playing song.
			// It will have changed if a song before it has
			// been deleted.
			//
			if (m_file_idx_selected < m_file_idx_playing)
			{
				m_file_idx_playing--;
			}
		}
		break;

	case 't':
	case 'T':
		// Toggle time display elapsed/remaining
		//
		m_time_show_remaining = !m_time_show_remaining;
		break;

	case 'r':
	case 'R':
		// Toggle RANDOM play on/off
		//
		m_play_random = !m_play_random;

		if (m_play_random)
		{
			// Seed the random number generator
			//
			struct timeval t;

			(int) gettimeofday(&t, NULL);

			srandom(t.tv_usec);
		}
		break;

	case 'l':
	case 'L':
		// Toggle LOOP play on/off
		//
		m_play_loop = !m_play_loop;
		break;

	case 'i':
	case 'I':
		// Toggle song mp3 info
		//
		m_info_show = !m_info_show;
		break;

	case '.':
		// Set file list to all files in the same
		// dir as currently selected file
		//
		song_select_dir_of_current_song(screen);
		break;

	case ',':
		// Make the playing file to the ONLY file in
		// the file list
		//
		song_set_file(screen, list_data(list_nth(m_file_list, m_file_idx_playing)));
		break;

	case '\t':
		// Toggle between filename and fullpath view
		//
		m_name_show_fullpath = !m_name_show_fullpath;
		song_set_display_name(screen, m_file_idx_playing);
		break;

	case KEY_SELECT:
		song_play(screen);
		break;

	case KEY_SONG_SELECT_PREVIOUS:
		if (-1 != m_file_idx_last_song)
		{
			m_file_idx_selected = m_file_idx_last_song;
			song_handle_key(screen, KEY_SELECT);
		}
		break;

	case KEY_DOWN:
		song_select_next(screen);
		break;

	case KEY_UP:
		song_select_prev(screen);
		break;

	case KEY_SHIFT_DOWN:
		song_select_next(screen);
		song_play(screen);
		break;

	case KEY_SHIFT_UP:
		song_select_prev(screen);
		song_play(screen);
		break;

	case KEY_SEARCH:
		// Pop up a new search screen
		//
		display_push(search_new(NULL));
		break;


	default:
		return KEY_NOT_HANDLED;
	}

	return KEY_HANDLED;
}


/******************************************************************************/

static void song_delete(screen_t * screen)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));

	// Free the song list by setting it to NULL.
	//
	song_set_file_list(screen, NULL, 0);

	// Free the screen object.
	//
	free(screen);

	song_screen_counter--;
	active_song_screen = NULL;
}

/******************************************************************************/

void song_file_list_add(screen_t * screen, file_entry_t * entry)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));
	assert(NULL != entry);

	m_file_list = list_add(m_file_list, entry);
}


/******************************************************************************/

void song_set_file(screen_t * screen, file_entry_t * entry)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));
	assert(NULL != entry);

	// Ignore the entry, if it is not a music file
	//
	if (!is_songfile_type(file_type(entry)))
		return;

	// It's save to free a NULL list
	//
	list_free(m_file_list);

	// add the new entry as first in the list
	//
	m_file_list = NULL;
	m_file_list = list_add(m_file_list, entry);
	m_file_idx_playing = 0;
	m_file_idx_selected = 0;
	m_file_idx_history_idx = 0;

	song_set_display_name(screen, 0);
}


/******************************************************************************/

void song_set_file_list(screen_t * screen, list_t *list, int current)
{
	list_t *actual_list = NULL;
	list_t *iter;

	// We count the current index while building the actual list.  We could
	// use list_index_of(entry) to figure out the index of entry, but it
	// would be much slower.
	//
	int list_current = 0;
	int actual_current = 0;


	assert((NULL != screen) && (TYPE_SONG == screen->type));

	// Create the actual file list and filter out any non-music files.
	//
	for (iter = list_first(list); iter; iter = list_next(iter))
	{
		file_entry_t *entry = list_data(iter);

		// Skip non-music files
		//
		if (!is_songfile_type(file_type(entry)))
		{
			list_current++;
			continue;
		}

		// Build up the actual list
		//
		actual_list = list_add(actual_list, entry);

		// Adjust the current index.  Some entries might be missing
		// before it.
		//
		if (list_current == current)
		{
			current = actual_current;
		}

		actual_current++;
		list_current++;
	}
	// delete the original list, we don't need it anymore
	//
	list_free(list);


	// Set the new file list
	//
	list_free(m_file_list);

	m_file_list = actual_list;
	m_file_idx_playing = current;
	m_file_idx_selected = current;
	m_file_idx_history_idx = 0;

	song_set_display_name(screen, current);
}


/******************************************************************************/

void song_play(screen_t * screen)
{
	file_entry_t *entry;
	const char *url;
	int type;

	assert((NULL != screen) && (TYPE_SONG == screen->type));

	entry = list_data(list_nth(m_file_list, m_file_idx_selected));

	// Only start playing, if we really have a mp3/ogg file or an URL
	//
	type = file_type(entry);

	switch (type)
	{
	case FILE_TYPE_MP3:
	case FILE_TYPE_OGG:
	case FILE_TYPE_FLAC:
		// cache the name for display purposes
		//
		song_set_display_name(screen, m_file_idx_selected);

		// And start playing
		//
		player_play(entry->fullpath, type);
		break;

	case FILE_TYPE_URL:
		// Get the url out of the file and play it through mpg123.
		//
		url = url_from_file(entry->fullpath);

		// We use type FILE_TYPE_MP3 so the player uses mpg123 when playing
		// the url and not ogg123
		//
		player_play(url, FILE_TYPE_MP3);
		break;

	default:
		assert(0);
	}

	// remember the currently playing song as the last song
	//
	m_file_idx_last_song = m_file_idx_playing;

	// Update playing song index
	//
	m_file_idx_playing = m_file_idx_selected;
	m_current_mode = SONG_MODE_PLAY;
}


/******************************************************************************/

screen_t *song_new()
{
	screen_t *screen;

	// NOTE: We actually allocate a song_t object, but we handle it with
	// its parent 'class' (screen_t) pointer.
	//
	screen = (screen_t *) malloc(sizeof(song_t));
	assert(NULL != screen);

	bzero(screen, sizeof(song_t));
	screen->type		= TYPE_SONG;
	screen->update		= song_update;
	screen->delete		= song_delete;
	screen->handle_key	= song_handle_key;
	screen->ctx		= &m_ctx;

	m_file_idx_last_song = -1;	// no last song

	song_screen_counter++;
	return screen;
}


/******************************************************************************/

void song_suspend(screen_t * screen)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));

	// When suspending a song, we remember the currently played frame
	//
	// When resuming, we can re-start from there.
	//
	m_current_frame = player_current_frame();

	// Now stop playing the current song
	//
	player_stop();
}


/******************************************************************************/

void song_resume(screen_t * screen)
{
	assert((NULL != screen) && (TYPE_SONG == screen->type));

	// Before we resume playing, we check, if there is another song
	// playing, in which case we will suspend it. This can happen when
	// switching between sessions.
	//
	if (song_active_screen() && (song_active_screen() != screen))
	{
		song_suspend(song_active_screen());
	}


	// Reset the saved state of the current song.
	//
	// Start playing the current song and then immediately set the
	// frame...
	//
	song_play(screen);
	player_goto_frame(m_current_frame);
}
